from recompy.FunkSVD import FunkSVD
import numpy as np
import pkg_resources


def load_movie_data():

    DATA_PATH = pkg_resources.resource_filename('recompy', 'data/')

    return np.genfromtxt(DATA_PATH + 'movielens100k.csv',
                         delimiter=',', skip_header=1)



# import library files
from recompy.Test import Test
from recompy.Initializer import initializer
from recompy.Similarities import Similarities

# import other libraries
import numpy as np
import copy
import bisect


class FunkSVD():
    def __init__(self):
        # Initialize default hyperparameters
        self.set_hyperparameters()

    def set_hyperparameters(self, initialization_method='random', max_epoch=5, n_latent=10, learning_rate=0.01, regularization=0.1, early_stopping=False, init_mean=0, init_std=1):
        self.initialization_method = initialization_method
        self.max_epoch = max_epoch
        self.n_latent = n_latent
        self.learning_rate = learning_rate
        self.regularization = regularization
        self.early_stopping = early_stopping
        self.init_mean = init_mean
        self.init_std = init_std

        self.min_train_error = np.inf
        self.min_test_error = np.inf

    def __set_data(self, data, test_portion):

        # get distinct users, items and user_existing_ratings, items_existing_users
        self.user_existing_ratings = {}
        self.items_rated_by_users = {}
        self.user_ids = []
        self.item_ids = []

        np.random.shuffle(data)

        # variables for train and test split
        user_dictionary = {}
        item_dictionary = {}
        self.train_data = []
        self.test_data = []

        for user, item, score in data:
            # Unique users and items

            try:
                user = int(user)
            except:
                pass
            try:
                item = int(item)
            except:
                pass

            user = str(user)
            item = str(item)
            score = float(score)

            if user not in self.user_existing_ratings:
                self.user_ids.append(user)
            if item not in self.items_rated_by_users:
                self.item_ids.append(item)

            self.items_rated_by_users.setdefault(item, []).append(user)
            self.user_existing_ratings.setdefault(user, []).append(item)

            ratio = len(self.test_data) / (len(self.train_data)+0.001)

            if self.test_split:
                # train and test set
                user_dictionary.setdefault(user, 0)
                item_dictionary.setdefault(item, 0)

                if user_dictionary[user] * test_portion >= 1 and item_dictionary[item] * test_portion >= 1 and ratio <= test_portion+0.02:

                    self.test_data.append([user, item, score])

                    user_dictionary[user] -= 1
                    item_dictionary[item] -= 1

                else:
                    self.train_data.append([user, item, score])

                    user_dictionary[user] += 1
                    item_dictionary[item] += 1
            else:
                self.train_data.append([user, item, score])

        print('Your data has {} distinct users and {} distinct items.'.format(
            len(self.user_ids), len(self.item_ids)))

        if len(self.test_data) < 1 and self.test_split:
            self.test_split = False
            self.early_stopping = False
            print("Training set doesn't have enough data for given test portion.")

        if self.test_split:

            print('Your data has been split into train and test set.')
            print('Length of training set is {}. Length of Test set is {}'.format(
                len(self.train_data), len(self.test_data)))
        else:

            print('Your data has no test set.')
            print('Length of training set is {}'.format(len(self.train_data)))

    def fit(self, data, test_split=True, test_portion=0.1, search_parameter_space=False):

        # Set train_data, test_data, user_ids etc. if search parameter is False
        # If True, this lets us search parameter space with the same train-test split
        if not search_parameter_space:

            self.test_split = test_split
            self.__set_data(data, test_portion)

        # Initialization
        print('Initializing features for Users and Items...')
        initial = initializer(self.user_ids, self.item_ids, self.initialization_method,
                              self.n_latent, self.init_mean, self.init_std)

        self.user_features, self.item_features = initial.initialize_latent_vectors()

        # Training
        print('Starting training...')
        error_counter = 0
        for epoch in range(self.max_epoch):

            # updating user and item features
            for user, item, rating in self.train_data:

                error = rating - \
                    np.dot(self.user_features[user], self.item_features[item])
                # Use temp to update each item and user feature in sync.
                temp = self.user_features[user]

                # Update user and item feature for each user, item and rating pair
                self.user_features[user] += self.learning_rate * \
                    (error * self.item_features[item] -
                     self.regularization * self.user_features[user])
                self.item_features[item] += self.learning_rate * \
                    (error * temp - self.regularization *
                     self.item_features[item])

            # Calculate errors
            error_counter += 1
            train_error = Test.rmse_error(
                self.train_data, self.user_features, self.item_features)

            # Show error to Client
            if self.test_split:
                test_error = Test.rmse_error(
                    self.test_data, self.user_features, self.item_features)
                print('Epoch Number: {}/{} Training RMSE: {:.2f} Test RMSE: {}'.format(epoch+1, self.max_epoch,
                                                                                       train_error, test_error))

            else:
                print('Epoch Number: {}/{} Training RMSE: {:.2f}'.format(epoch+1, self.max_epoch,
                                                                         train_error))

            # Save best features depending on test_error
            if self.test_split and test_error < self.min_test_error:
                self.min_test_error = test_error
                self.best_user_features = copy.deepcopy(self.user_features)
                self.best_item_features = copy.deepcopy(self.item_features)

                error_counter = 0
            # Save best features if test data is False
            elif not self.test_split and train_error < self.min_train_error:
                self.min_train_error = train_error
                self.best_user_features = copy.deepcopy(self.user_features)
                self.best_item_features = copy.deepcopy(self.item_features)

            # Break if test_error didn't improve for the last n rounds and early stopping is true
            if self.early_stopping and error_counter >= self.early_stopping:

                print("Test error didn't get lower for the last {} epochs. Training is stopped.".format(
                    error_counter))
                print('Best test error is: {:.2f}. Best features are saved.'.format(
                    self.min_test_error))
                break

        print('Training has ended...')
        self.user_features = copy.deepcopy(self.best_user_features)
        self.item_features = copy.deepcopy(self.best_item_features)

    def get_recommendation_for_existing_user(self, user_id, howMany=10):
        result_list = []
        # this might be more effective using matrix multiplication
        for item in self.item_ids:
            # if user did not already rate the item
            if item not in self.user_existing_ratings[user_id]:
                prediction = np.dot(
                    self.user_features[user_id], self.item_features[item])
                bisect.insort(result_list, [prediction, item])

        return [x[1] for x in result_list[::-1][0:howMany]]

    def get_recommendation_for_new_user(self, user_ratings,
                                        similarity_measure='mean_squared_difference', howManyUsers=3, howManyItems=5):

        # Get user predictions on same movies
        user_predictions = self.__user_prediction_for_same_movies(user_ratings)
        # Find most most similar user_ids
        user_ids = Similarities.get_most_similar_users(
            user_ratings, user_predictions, similarity_measure, howManyUsers)

        result_list = []
        # get user features for users who are most similar to given new user
        for user in user_ids:
            for item, item_feature in self.item_features.items():
                # predict ratings for most similar users
                prediction = np.dot(
                    self.user_features[user], item_feature)
                bisect.insort(result_list, [prediction, item])

        return [x[1] for x in result_list[::-1][:howManyItems]]

    def get_similar_products(self, product_id, howMany=10):

        result_list = []
        product_features = self.item_features[product_id]

        for item in self.item_ids:

            if item == product_id:
                continue
            # add cosine sim function from similarites
            cos_sim = Similarities.cosine_similarity(
                self.item_features[item], product_features)

            bisect.insort(result_list, [cos_sim, item])

        return [x[1] for x in result_list[::-1][0:howMany]]

    def __user_prediction_for_same_movies(self, user_ratings):
        result = {}
        for key in user_ratings:
            if key not in self.item_features:
                continue

            for user in self.user_features:
                result.setdefault(user, []).append(
                    np.dot(self.user_features[user], self.item_features[key]))

        return result



import numpy as np
from numpy.linalg import norm


class Similarities():

    @staticmethod
    def get_most_similar_users(user_ratings, user_predictions, similarity_measure, howMany):
        similarities = []

        for user, ratings in user_predictions.items():
            if similarity_measure == 'cosine_similarity':
                similarity = Similarities.cosine_similarity(
                    list(user_ratings.values()), ratings)
            elif similarity_measure == 'pearson_correlation':
                similarity = Similarities.pearson_correlation(
                    list(user_ratings.values()), ratings)
            elif similarity_measure == 'adjusted_cosine_similarity':
                similarity = Similarities.adjusted_cosine_similarity(
                    list(user_ratings.values()), ratings)
            elif similarity_measure == 'weighted_cosine_similarity':
                similarity = Similarities.weighted_cosine_similarity(
                    list(user_ratings.values()), ratings)
            elif similarity_measure == 'constrained_pearson_correlation':
                similarity = Similarities.weighted_cosine_similarity(
                    list(user_ratings.values()), ratings)
            elif similarity_measure == 'mean_squared_difference':
                similarity = Similarities.mean_squared_difference(
                    list(user_ratings.values()), ratings)
            elif similarity_measure == 'constrained_pearson_correlation':
                similarity = Similarities.constrained_pearson_correlation(
                    list(user_ratings.values()), ratings)

            similarities.append([user, similarity])

        similarities.sort(reverse=True, key=lambda x: x[1])

        return [each[0] for each in similarities[:howMany]]

    @staticmethod
    def cosine_similarity(a, b):

        return np.dot(a, b)/(norm(a)*norm(b))

    @staticmethod
    def pearson_correlation(a, b):
        corr = np.corrcoef(a, b)[0, 1]

        return corr

    @staticmethod
    def weighted_cosine_similarity(a, b):
        shared_item_count = len(a)
        cosine_similarity = np.dot(a, b)/(norm(a)*norm(b))
        weighted_cosine_similarity = cosine_similarity * \
            (1 / (1+np.exp(-1*shared_item_count)))

        return weighted_cosine_similarity

    @staticmethod
    def adjusted_cosine_similarity(a, b):
        mean_response = sum(sum(a, b)) / (2*len(a))
        a = a - mean_response
        b = b - mean_response

        return np.dot(a, b)/(norm(a)*norm(b))

    @staticmethod
    def mean_squared_difference(a, b):
        summation = 0
        n = len(a)
        for i in range(0, n):
            difference = a[i] - b[i]
            squared_difference = difference**2
            summation = summation + squared_difference
        MSE = summation/n

        return 1/MSE

    @staticmethod
    def constrained_pearson_correlation(a, b):
        median_a = np.median(a)
        median_b = np.median(b)
        nominator = np.dot((a - median_a), (b - median_b))
        denominator1 = np.sqrt(np.dot((a - median_a), (a - median_a)))
        denominator2 = np.sqrt(np.dot((b - median_b), (b - median_b)))
        denominator = denominator1 * denominator2
        cpc = nominator / denominator

        return cpc



import numpy as np


class initializer():
    def __init__(self, user_ids, item_ids, method, n_latent, mean, std):
        self.__mean = mean
        self.__std = std
        self.__n_latent = n_latent
        self.__user_ids = user_ids
        self.__item_ids = item_ids
        self.initalization_method = method

    @staticmethod
    def random_initializer(n_latent):
        return np.random.random_sample((n_latent,))

    @staticmethod
    def normal_initializer(mean, std, n_latent):
        return np.random.normal(mean, std, (n_latent,))

    @staticmethod
    def he_initializer(n_latent):
        return np.random.randn(n_latent) * np.sqrt(2/n_latent)

    def initialize_latent_vectors(self, initalization_method='he'):
        # Initialize item and user features as disctionary
        try:
                # if not valid method, return exception
            ['random', 'normal', 'he'].index(self.initalization_method)

        except ValueError:
            print(
                "Invalid initializer method. Choose among ['random','normal','he']")

        else:
            self.item_features = {}
            self.user_features = {}

            for user in self.__user_ids:
                # Generate random number for each user
                self.user_features.setdefault(
                    user, self.generate_random_feature())

            for item in self.__item_ids:
                # Generate random number for each item
                self.item_features.setdefault(
                    item, self.generate_random_feature())

            return self.user_features, self.item_features

    def generate_random_feature(self):

        # Generate features depending on given method
        if self.initalization_method == 'random':
            return self.random_initializer(self.__n_latent)

        elif self.initalization_method == 'normal':
            return self.normal_initializer(self.__mean, self.__std, self.__n_latent)

        elif self.initalization_method == 'he':
            return self.he_initializer(self.__n_latent)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="recompy",
    version="1.0.1",
    author="Can Bulguoglu, Oguz Kaplan, Onur Boyacı, Emre Yuksel",
    author_email="canbulguoglu@gmail.com",
    description="A recommender library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/canbul/recompy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["numpy==1.18.4"],
    python_requires='>=3.5',
    include_package_data=True,
    package_data={
        '': ['*.csv']
    }
)



import numpy as np


class Test():

    @staticmethod
    def rmse_error(data, user_features, item_features):
        # Get RMSE for the given data (train or test)

        total_error = 0
        counter = 0

        for user, item, rating in data:
            if user not in user_features or item not in item_features:
                continue

            total_error += (rating -
                            np.dot(user_features[user], item_features[item]))**2
            counter += 1

        return np.sqrt(total_error/counter)

    @staticmethod
    def novelty(train_data, recommendation_list, user_n):

        novelty = 0

        for movie in recommendation_list:
            # Calculate novelty for each item in the recommendation list
            pop_item = train_data[train_data[:, 1] == movie].shape[0]
            novelty += 1 - (pop_item/user_n)

        novelty = novelty/len(recommendation_list)

        return novelty

    @staticmethod
    def precision_recall_at_k(test_data, user_features, item_features, threshold, k):
        user_true_pred = dict()
        precision_k = dict()
        recall_k = dict()

        for row in test_data:

            user_id = row[0]
            item_id = row[1]
            true_rating = row[2]

            # Check the user in the test set also in the training set
            if user_id not in user_features or item_id not in item_features:
                continue
            estimated_rating = np.dot(
                user_features[user_id], item_features[item_id])
            try:
                user_true_pred[user_id].append((true_rating, estimated_rating))

            except KeyError:
              # Create a dictionary with list as values
                user_true_pred[user_id] = []
                user_true_pred[user_id].append((true_rating, estimated_rating))

        for user_id, rating in user_true_pred.items():

            rating.sort(key=lambda x: x[1], reverse=True)

            # Number of recommended items at k
            recommended_c = sum((estimated >= threshold)
                                for (_, estimated) in rating[:k])

            # Number of relevant items
            relevant_c = sum((true >= threshold) for (true, _) in rating)

            # Number of relevant and recommended items in top k
            recommended_in_relevant = sum(((estimated >= threshold) and (true >= threshold))
                                          for (true, estimated) in rating[:k])

            # Precision at K
            precision_k[user_id] = recommended_in_relevant / \
                recommended_c if recommended_c != 0 else 0

            # Recall at K
            recall_k[user_id] = recommended_in_relevant / \
                relevant_c if relevant_c != 0 else 0

            # Precision and recall can then be averaged over all users
            precision = sum(
                prec for prec in precision_k.values()) / len(precision_k)
            recall = sum(rec for rec in recall_k.values()) / len(recall_k)

        print('Predicision@K: {}\nRecall@K: {}'.format(precision, recall))

        return [precision, recall]
